=== modified file 'bzrlib/atomicfile.py'
--- bzrlib/atomicfile.py	2005-09-23 09:25:16 +0000
+++ bzrlib/atomicfile.py	2005-09-29 20:34:25 +0000
@@ -88,6 +88,6 @@
 
 
     def __del__(self):
-        if not self.closed:
+        if hasattr(self, 'closed') and not self.closed:
             warn("%r leaked" % self)
         

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-09-29 12:20:42 +0000
+++ bzrlib/branch.py	2005-09-29 20:34:25 +0000
@@ -19,6 +19,7 @@
 import os
 import errno
 from warnings import warn
+from cStringIO import StringIO
 
 
 import bzrlib
@@ -29,14 +30,16 @@
 from bzrlib.errors import (BzrError, InvalidRevisionNumber, InvalidRevisionId,
                            NoSuchRevision, HistoryMissing, NotBranchError,
                            DivergedBranches, LockError, UnlistableStore,
-                           UnlistableBranch)
+                           UnlistableBranch, NoSuchFile)
 from bzrlib.textui import show_status
 from bzrlib.revision import Revision, validate_revision_id, is_ancestor
 from bzrlib.delta import compare_trees
 from bzrlib.tree import EmptyTree, RevisionTree
 from bzrlib.inventory import Inventory
 from bzrlib.weavestore import WeaveStore
-from bzrlib.store import copy_all, ImmutableStore
+from bzrlib.store import copy_all
+from bzrlib.store.compressed_text import CompressedTextStore
+from bzrlib.transport import Transport
 import bzrlib.xml5
 import bzrlib.ui
 
@@ -55,7 +58,6 @@
     # XXX: leave this here for about one release, then remove it
     raise NotImplementedError('find_branch() is not supported anymore, '
                               'please use one of the new branch constructors')
-
 def _relpath(base, path):
     """Return path relative to base, or raise exception.
 
@@ -83,38 +85,25 @@
     return os.sep.join(s)
         
 
-def find_branch_root(f=None):
-    """Find the branch root enclosing f, or pwd.
-
-    f may be a filename or a URL.
-
-    It is not necessary that f exists.
+def find_branch_root(t):
+    """Find the branch root enclosing the transport's base.
+
+    t is a Transport object.
+
+    It is not necessary that the base of t exists.
 
     Basically we keep looking up until we find the control directory or
     run into the root.  If there isn't one, raises NotBranchError.
     """
-    if f == None:
-        f = os.getcwd()
-    elif hasattr(os.path, 'realpath'):
-        f = os.path.realpath(f)
-    else:
-        f = os.path.abspath(f)
-    if not os.path.exists(f):
-        raise BzrError('%r does not exist' % f)
-        
-
-    orig_f = f
-
+    orig_base = t.base
     while True:
-        if os.path.exists(os.path.join(f, bzrlib.BZRDIR)):
-            return f
-        head, tail = os.path.split(f)
-        if head == f:
+        if t.has(bzrlib.BZRDIR):
+            return t
+        new_t = t.clone('..')
+        if new_t.base == t.base:
             # reached the root, whatever that may be
-            raise NotBranchError('%s is not in a branch' % orig_f)
-        f = head
-
-
+            raise NotBranchError('%s is not in a branch' % orig_base)
+        t = new_t
 
 
 ######################################################################
@@ -136,16 +125,13 @@
         """Open a branch which may be of an old format.
         
         Only local branches are supported."""
-        return LocalBranch(base, find_root=False, relax_version_check=True)
+        return _Branch(base, relax_version_check=True)
         
     @staticmethod
     def open(base):
         """Open an existing branch, rooted at 'base' (url)"""
-        if base and (base.startswith('http://') or base.startswith('https://')):
-            from bzrlib.remotebranch import RemoteBranch
-            return RemoteBranch(base, find_root=False)
-        else:
-            return LocalBranch(base, find_root=False)
+        t = bzrlib.transport.transport(base)
+        return _Branch(t)
 
     @staticmethod
     def open_containing(url):
@@ -153,20 +139,15 @@
         
         This probes for a branch at url, and searches upwards from there.
         """
-        if url and (url.startswith('http://') or url.startswith('https://')):
-            from bzrlib.remotebranch import RemoteBranch
-            return RemoteBranch(url)
-        else:
-            return LocalBranch(url)
+        t = bzrlib.transport.transport(url)
+        t = find_branch_root(t)
+        return _Branch(t)
 
     @staticmethod
     def initialize(base):
         """Create a new branch, rooted at 'base' (url)"""
-        if base and (base.startswith('http://') or base.startswith('https://')):
-            from bzrlib.remotebranch import RemoteBranch
-            return RemoteBranch(base, init=True)
-        else:
-            return LocalBranch(base, init=True)
+        t = bzrlib.transport.transport(base)
+        return _Branch(t, init=True)
 
     def setup_caching(self, cache_root):
         """Subclasses that care about caching should override this, and set
@@ -174,7 +155,7 @@
         """
 
 
-class LocalBranch(Branch):
+class _Branch(Branch):
     """A branch stored in the actual filesystem.
 
     Note that it's "local" in the context of the filesystem; it doesn't
@@ -225,19 +206,18 @@
         except UnlistableStore:
             raise UnlistableBranch(from_store)
 
-    def __init__(self, base, init=False, find_root=True,
+    def __init__(self, transport, init=False,
                  relax_version_check=False):
         """Create new branch object at a particular location.
 
-        base -- Base directory for the branch. May be a file:// url.
+        transport -- A Transport object, defining how to access files.
+                (If a string, transport.transport() will be used to
+                create a Transport object)
         
         init -- If True, create new control files in a previously
              unversioned directory.  If False, the branch must already
              be versioned.
 
-        find_root -- If true and init is false, find the root of the
-             existing branch containing base.
-
         relax_version_check -- If true, the usual check for the branch
             version is not applied.  This is intended only for
             upgrade/recovery type use; it's not guaranteed that
@@ -246,24 +226,15 @@
         In the test suite, creation of new trees is tested using the
         `ScratchBranch` class.
         """
+        assert isinstance(transport, Transport)
+        self._transport = transport
         if init:
-            self.base = os.path.realpath(base)
             self._make_control()
-        elif find_root:
-            self.base = find_branch_root(base)
-        else:
-            if base.startswith("file://"):
-                base = base[7:]
-            self.base = os.path.realpath(base)
-            if not isdir(self.controlfilename('.')):
-                raise NotBranchError('not a bzr branch: %s' % quotefn(base),
-                                     ['use "bzr init" to initialize a '
-                                      'new working tree'])
         self._check_format(relax_version_check)
         cfn = self.controlfilename
         if self._branch_format == 4:
-            self.inventory_store = ImmutableStore(cfn('inventory-store'))
-            self.text_store = ImmutableStore(cfn('text-store'))
+            self.inventory_store = CompressedTextStore(cfn('inventory-store'))
+            self.text_store = CompressedTextStore(cfn('text-store'))
         elif self._branch_format == 5:
             self.control_weaves = WeaveStore(cfn([]))
             self.weave_store = WeaveStore(cfn('weaves'))
@@ -271,11 +242,10 @@
                 # FIXME: Unify with make_control_files
                 self.control_weaves.put_empty_weave('inventory')
                 self.control_weaves.put_empty_weave('ancestry')
-        self.revision_store = ImmutableStore(cfn('revision-store'))
-
+        self.revision_store = CompressedTextStore(cfn('revision-store'))
 
     def __str__(self):
-        return '%s(%r)' % (self.__class__.__name__, self.base)
+        return '%s(%r)' % (self.__class__.__name__, self._transport.base)
 
 
     __repr__ = __str__
@@ -288,16 +258,39 @@
             warn("branch %r was not explicitly unlocked" % self)
             self._lock.unlock()
 
+        # TODO: It might be best to do this somewhere else,
+        # but it is nice for a Branch object to automatically
+        # cache it's information.
+        # Alternatively, we could have the Transport objects cache requests
+        # See the earlier discussion about how major objects (like Branch)
+        # should never expect their __del__ function to run.
+        if hasattr(self, 'cache_root') and self.cache_root is not None:
+            try:
+                import shutil
+                shutil.rmtree(self.cache_root)
+            except:
+                pass
+            self.cache_root = None
+
+    def _get_base(self):
+        if self._transport:
+            return self._transport.base
+        return None
+
+    base = property(_get_base)
+
+
     def lock_write(self):
+        # TODO: Upgrade locking to support using a Transport,
+        # and potentially a remote locking protocol
         if self._lock_mode:
             if self._lock_mode != 'w':
                 raise LockError("can't upgrade to a write lock from %r" %
                                 self._lock_mode)
             self._lock_count += 1
         else:
-            from bzrlib.lock import WriteLock
-
-            self._lock = WriteLock(self.controlfilename('branch-lock'))
+            self._lock = self._transport.lock_write(
+                    self._rel_controlfilename('branch-lock'))
             self._lock_mode = 'w'
             self._lock_count = 1
 
@@ -308,9 +301,8 @@
                    "invalid lock mode %r" % self._lock_mode
             self._lock_count += 1
         else:
-            from bzrlib.lock import ReadLock
-
-            self._lock = ReadLock(self.controlfilename('branch-lock'))
+            self._lock = self._transport.lock_read(
+                    self._rel_controlfilename('branch-lock'))
             self._lock_mode = 'r'
             self._lock_count = 1
                         
@@ -327,19 +319,23 @@
 
     def abspath(self, name):
         """Return absolute filename for something in the branch"""
-        return os.path.join(self.base, name)
+        return self._transport.abspath(name)
 
     def relpath(self, path):
         """Return path relative to this branch of something inside it.
 
         Raises an error if path is not in this branch."""
-        return _relpath(self.base, path)
+        return self._transport.relpath(path)
+
+
+    def _rel_controlfilename(self, file_or_path):
+        if isinstance(file_or_path, basestring):
+            file_or_path = [file_or_path]
+        return [bzrlib.BZRDIR] + file_or_path
 
     def controlfilename(self, file_or_path):
         """Return location relative to branch."""
-        if isinstance(file_or_path, basestring):
-            file_or_path = [file_or_path]
-        return os.path.join(self.base, bzrlib.BZRDIR, *file_or_path)
+        return self._transport.abspath(self._rel_controlfilename(file_or_path))
 
 
     def controlfile(self, file_or_path, mode='r'):
@@ -353,42 +349,76 @@
         Controlfiles should almost never be opened in write mode but
         rather should be atomically copied and replaced using atomicfile.
         """
-
-        fn = self.controlfilename(file_or_path)
-
-        if mode == 'rb' or mode == 'wb':
-            return file(fn, mode)
-        elif mode == 'r' or mode == 'w':
-            # open in binary mode anyhow so there's no newline translation;
-            # codecs uses line buffering by default; don't want that.
-            import codecs
-            return codecs.open(fn, mode + 'b', 'utf-8',
-                               buffering=60000)
+        import codecs
+
+        relpath = self._rel_controlfilename(file_or_path)
+        #TODO: codecs.open() buffers linewise, so it was overloaded with
+        # a much larger buffer, do we need to do the same for getreader/getwriter?
+        if mode == 'rb': 
+            return self._transport.get(relpath)
+        elif mode == 'wb':
+            raise BzrError("Branch.controlfile(mode='wb') is not supported, use put_controlfiles")
+        elif mode == 'r':
+            return codecs.getreader('utf-8')(self._transport.get(relpath), errors='replace')
+        elif mode == 'w':
+            raise BzrError("Branch.controlfile(mode='w') is not supported, use put_controlfiles")
         else:
             raise BzrError("invalid controlfile mode %r" % mode)
 
+    def put_controlfile(self, path, f, encode=True):
+        """Write an entry as a controlfile.
+
+        :param path: The path to put the file, relative to the .bzr control
+                     directory
+        :param f: A file-like or string object whose contents should be copied.
+        :param encode:  If true, encode the contents as utf-8
+        """
+        self.put_controlfiles([(path, f)], encode=encode)
+
+    def put_controlfiles(self, files, encode=True):
+        """Write several entries as controlfiles.
+
+        :param files: A list of [(path, file)] pairs, where the path is the directory
+                      underneath the bzr control directory
+        :param encode:  If true, encode the contents as utf-8
+        """
+        import codecs
+        ctrl_files = []
+        for path, f in files:
+            if encode:
+                if isinstance(f, basestring):
+                    f = f.encode('utf-8', 'replace')
+                else:
+                    f = codecs.getwriter('utf-8')(f, errors='replace')
+            path = self._rel_controlfilename(path)
+            ctrl_files.append((path, f))
+        self._transport.put_multi(ctrl_files)
+
     def _make_control(self):
-        os.mkdir(self.controlfilename([]))
-        self.controlfile('README', 'w').write(
-            "This is a Bazaar-NG control directory.\n"
-            "Do not change any files in this directory.\n")
-        self.controlfile('branch-format', 'w').write(BZR_BRANCH_FORMAT_5)
-        for d in ('text-store', 'revision-store',
-                  'weaves'):
-            os.mkdir(self.controlfilename(d))
-        for f in ('revision-history',
-                  'branch-name',
-                  'branch-lock',
-                  'pending-merges'):
-            self.controlfile(f, 'w').write('')
-        mutter('created control directory in ' + self.base)
-
+        from bzrlib.inventory import Inventory
+        
+        # Create an empty inventory
+        sio = StringIO()
         # if we want per-tree root ids then this is the place to set
         # them; they're not needed for now and so ommitted for
         # simplicity.
-        f = self.controlfile('inventory','w')
-        bzrlib.xml5.serializer_v5.write_inventory(Inventory(), f)
+        bzrlib.xml5.serializer_v5.write_inventory(Inventory(), sio)
 
+        dirs = [[], 'revision-store', 'weaves']
+        files = [('README', 
+            "This is a Bazaar-NG control directory.\n"
+            "Do not change any files in this directory.\n"),
+            ('branch-format', BZR_BRANCH_FORMAT_5),
+            ('revision-history', ''),
+            ('branch-name', ''),
+            ('branch-lock', ''),
+            ('pending-merges', ''),
+            ('inventory', sio.getvalue())
+        ]
+        cfn = self._rel_controlfilename
+        self._transport.mkdir_multi([cfn(d) for d in dirs])
+        self.put_controlfiles(files)
+        mutter('created control directory in ' + self._transport.base)
 
     def _check_format(self, relax_version_check):
         """Check this branch format is supported.
@@ -401,11 +431,8 @@
         """
         try:
             fmt = self.controlfile('branch-format', 'r').read()
-        except IOError, e:
-            if e.errno == errno.ENOENT:
-                raise NotBranchError(self.base)
-            else:
-                raise
+        except NoSuchFile:
+            raise NotBranchError(self.base)
 
         if fmt == BZR_BRANCH_FORMAT_5:
             self._branch_format = 5
@@ -416,7 +443,8 @@
             and self._branch_format != 5):
             raise BzrError('sorry, branch format %r not supported' % fmt,
                            ['use a different bzr version',
-                            'or remove the .bzr directory and "bzr init" again'])
+                            'or remove the .bzr directory'
+                            ' and "bzr init" again'])
 
     def get_root_id(self):
         """Return the id of this branches root"""
@@ -453,16 +481,14 @@
         That is to say, the inventory describing changes underway, that
         will be committed to the next revision.
         """
-        from bzrlib.atomicfile import AtomicFile
-        
+        from cStringIO import StringIO
         self.lock_write()
         try:
-            f = AtomicFile(self.controlfilename('inventory'), 'wb')
-            try:
-                bzrlib.xml5.serializer_v5.write_inventory(inv, f)
-                f.commit()
-            finally:
-                f.close()
+            sio = StringIO()
+            bzrlib.xml5.serializer_v5.write_inventory(inv, sio)
+            sio.seek(0)
+            # Transport handles atomicity
+            self.put_controlfile('inventory', sio)
         finally:
             self.unlock()
         
@@ -633,21 +659,17 @@
 
 
     def append_revision(self, *revision_ids):
-        from bzrlib.atomicfile import AtomicFile
-
         for revision_id in revision_ids:
             mutter("add {%s} to revision-history" % revision_id)
 
         rev_history = self.revision_history()
         rev_history.extend(revision_ids)
 
-        f = AtomicFile(self.controlfilename('revision-history'))
+        self.lock_write()
         try:
-            for rev_id in rev_history:
-                print >>f, rev_id
-            f.commit()
+            self.put_controlfile('revision-history', '\n'.join(rev_history))
         finally:
-            f.close()
+            self.unlock()
 
 
     def has_revision(self, revision_id):
@@ -714,7 +736,6 @@
 
         return compare_trees(old_tree, new_tree)
 
-
     def get_revision_sha1(self, revision_id):
         """Hash the stored value of a revision, and return it."""
         return bzrlib.osutils.sha_file(self.get_revision_xml_file(revision_id))
@@ -951,7 +972,8 @@
     def working_tree(self):
         """Return a `Tree` for the working copy."""
         from bzrlib.workingtree import WorkingTree
-        return WorkingTree(self.base, self.read_working_inventory())
+        # TODO: In the future, WorkingTree should utilize Transport
+        return WorkingTree(self._transport.base, self.read_working_inventory())
 
 
     def basis_tree(self):
@@ -1131,8 +1153,8 @@
         These are revisions that have been merged into the working
         directory but not yet committed.
         """
-        cfn = self.controlfilename('pending-merges')
-        if not os.path.exists(cfn):
+        cfn = self._rel_controlfilename('pending-merges')
+        if not self._transport.has(cfn):
             return []
         p = []
         for l in self.controlfile('pending-merges', 'r').readlines():
@@ -1140,28 +1162,26 @@
         return p
 
 
-    def add_pending_merge(self, revision_id):
-        validate_revision_id(revision_id)
+    def add_pending_merge(self, *revision_ids):
         # TODO: Perhaps should check at this point that the
         # history of the revision is actually present?
+        for rev_id in revision_ids:
+            validate_revision_id(rev_id)
+
         p = self.pending_merges()
-        if revision_id in p:
-            return
-        p.append(revision_id)
-        self.set_pending_merges(p)
-
+        updated = False
+        for rev_id in revision_ids:
+            if rev_id in p:
+                continue
+            p.append(rev_id)
+            updated = True
+        if updated:
+            self.set_pending_merges(p)
 
     def set_pending_merges(self, rev_list):
-        from bzrlib.atomicfile import AtomicFile
         self.lock_write()
         try:
-            f = AtomicFile(self.controlfilename('pending-merges'))
-            try:
-                for l in rev_list:
-                    print >>f, l
-                f.commit()
-            finally:
-                f.close()
+            self.put_controlfile('pending-merges', '\n'.join(rev_list))
         finally:
             self.unlock()
 
@@ -1218,7 +1238,7 @@
         
 
 
-class ScratchBranch(LocalBranch):
+class ScratchBranch(_Branch):
     """Special test class: a branch that cleans up after itself.
 
     >>> b = ScratchBranch()
@@ -1241,12 +1261,12 @@
         if base is None:
             base = mkdtemp()
             init = True
-        LocalBranch.__init__(self, base, init=init)
+        _Branch.__init__(self, base, init=init)
         for d in dirs:
-            os.mkdir(self.abspath(d))
+            self._transport.mkdir(d)
             
         for f in files:
-            file(os.path.join(self.base, f), 'w').write('content of %s' % f)
+            self._transport.put(f, 'content of %s' % f)
 
 
     def clone(self):
@@ -1289,7 +1309,7 @@
                 for name in files:
                     os.chmod(os.path.join(root, name), 0700)
             rmtree(self.base)
-        self.base = None
+        self._transport = None
 
     
 

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2005-09-29 12:20:42 +0000
+++ bzrlib/builtins.py	2005-09-29 20:34:25 +0000
@@ -1376,11 +1376,9 @@
             # should not change the parent
             b.set_parent(remote)
         br_remote = Branch.open_containing(remote)
-
         return show_missing(b, br_remote, verbose=verbose, quiet=quiet)
 
 
-
 class cmd_plugins(Command):
     """List plugins"""
     hidden = True

=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2005-09-28 09:35:50 +0000
+++ bzrlib/commands.py	2005-09-29 20:34:25 +0000
@@ -571,6 +571,8 @@
     --profile
         Run under the Python profiler.
     """
+    # Load all of the transport methods
+    import bzrlib.transport.local, bzrlib.transport.http
     
     argv = [a.decode(bzrlib.user_encoding) for a in argv]
 
@@ -662,6 +664,5 @@
             bzrlib.trace.log_exception()
             return 2
 
-
 if __name__ == '__main__':
     sys.exit(main(sys.argv))

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-09-29 00:24:44 +0000
+++ bzrlib/errors.py	2005-09-29 20:34:25 +0000
@@ -167,3 +167,54 @@
 
 
 from bzrlib.weave import WeaveError
+
+class TransportError(BzrError):
+    """All errors thrown by Transport implementations should derive
+    from this class.
+    """
+    def __init__(self, msg=None, orig_error=None):
+        if msg is None and orig_error is not None:
+            msg = str(orig_error)
+        BzrError.__init__(self, msg)
+        self.msg = msg
+        self.orig_error = orig_error
+
+# A set of semi-meaningful errors which can be thrown
+class TransportNotPossible(TransportError):
+    """This is for transports where a specific function is explicitly not
+    possible. Such as pushing files to an HTTP server.
+    """
+    pass
+
+class NonRelativePath(TransportError):
+    """An absolute path was supplied, that could not be decoded into
+    a relative path.
+    """
+    pass
+
+class NoSuchFile(TransportError, IOError):
+    """A get() was issued for a file that doesn't exist."""
+    def __init__(self, msg=None, orig_error=None):
+        import errno
+        TransportError.__init__(self, msg=msg, orig_error=orig_error)
+        IOError.__init__(self, errno.ENOENT, self.msg)
+
+class FileExists(TransportError, OSError):
+    """An operation was attempted, which would overwrite an entry,
+    but overwritting is not supported.
+
+    mkdir() can throw this, but put() just overwites existing files.
+    """
+    def __init__(self, msg=None, orig_error=None):
+        import errno
+        TransportError.__init__(self, msg=msg, orig_error=orig_error)
+        OSError.__init__(self, errno.EEXIST, self.msg)
+
+class PermissionDenied(TransportError):
+    """An operation cannot succeed because of a lack of permissions."""
+    pass
+
+class ConnectionReset(TransportError):
+    """The connection has been closed."""
+    pass
+

=== modified file 'bzrlib/meta_store.py'
--- bzrlib/meta_store.py	2005-09-26 08:56:15 +0000
+++ bzrlib/meta_store.py	2005-09-29 20:34:25 +0000
@@ -15,17 +15,18 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 from bzrlib.trace import mutter
-from bzrlib.store import ImmutableStore
-
-
-class CachedStore:
+from bzrlib.store import Store
+from bzrlib.store.compressed_text import CompressedTextStore
+from bzrlib.transport.local import LocalTransport
+
+class CachedStore(Store):
     """A store that caches data locally, to avoid repeated downloads.
     The precacache method should be used to avoid server round-trips for
     every piece of data.
     """
     def __init__(self, store, cache_dir):
         self.source_store = store
-        self.cache_store = ImmutableStore(cache_dir)
+        self.cache_store = CompressedTextStore(LocalTransport(cache_dir))
 
     def __getitem__(self, id):
         mutter("Cache add %s" % id)
@@ -33,6 +34,27 @@
             self.cache_store.add(self.source_store[id], id)
         return self.cache_store[id]
 
+    def __contains__(self, fileid):
+        if fileid in self.cache_store:
+            return True
+        if fileid in self.source_store:
+            # We could copy at this time
+            return True
+        return False
+
+    def get(self, fileids, permit_failure=False, pb=None):
+        fileids = list(fileids)
+        hasids = self.cache_store.has(fileids)
+        needs = set()
+        for has, fileid in zip(hasids, fileids):
+            if not has:
+                needs.add(fileid)
+        if needs:
+            self.cache_store.copy_multi(self.source_store, needs,
+                    permit_failure=permit_failure)
+        return self.cache_store.get(fileids,
+                permit_failure=permit_failure, pb=pb)
+
     def prefetch(self, ids):
         """Copy a series of ids into the cache, before they are used.
         For remote stores that support pipelining or async downloads, this can
@@ -43,3 +65,4 @@
         mutter("Prefetch of ids %s" % ",".join(ids))
         self.cache_store.copy_multi(self.source_store, ids, 
                                     permit_failure=True)
+

=== removed file 'bzrlib/remotebranch.py'
--- bzrlib/remotebranch.py	2005-09-29 02:01:49 +0000
+++ bzrlib/remotebranch.py	1970-01-01 00:00:00 +0000
@@ -1,217 +0,0 @@
-# Copyright (C) 2005 Canonical Ltd
-
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
-# the Free Software Foundation; either version 2 of the License, or
-# (at your option) any later version.
-
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
-
-# You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-
-
-"""Proxy object for access to remote branches.
-
-At the moment remote branches are only for HTTP and only for read
-access.
-"""
-
-
-import gzip
-from cStringIO import StringIO
-import os
-import urllib2
-import urlparse
-
-from bzrlib.errors import BzrError, BzrCheckError
-from bzrlib.branch import Branch, LocalBranch, BZR_BRANCH_FORMAT_5
-from bzrlib.trace import mutter
-from bzrlib.weavestore import WeaveStore
-from bzrlib.xml5 import serializer_v5
-
-
-# velocitynet.com.au transparently proxies connections and thereby
-# breaks keep-alive -- sucks!
-
-
-ENABLE_URLGRABBER = False
-
-from bzrlib.errors import BzrError, NoSuchRevision
-
-class GetFailed(BzrError):
-    def __init__(self, url, status):
-        BzrError.__init__(self, "Get %s failed with status %s" % (url, status))
-        self.url = url
-        self.status = status
-
-if ENABLE_URLGRABBER:
-    import util.urlgrabber
-    import util.urlgrabber.keepalive
-    util.urlgrabber.keepalive.DEBUG = 0
-    def get_url(path, compressed=False):
-        try:
-            url = path
-            if compressed:
-                url += '.gz'
-            mutter("grab url %s" % url)
-            url_f = util.urlgrabber.urlopen(url, keepalive=1, close_connection=0)
-            if url_f.status != 200:
-                raise GetFailed(url, url_f.status)
-            if not compressed:
-                return url_f
-            else:
-                return gzip.GzipFile(fileobj=StringIO(url_f.read()))
-        except urllib2.URLError, e:
-            raise BzrError("remote fetch failed: %r: %s" % (url, e))
-else:
-    def get_url(url, compressed=False):
-        import urllib2
-        if compressed:
-            url += '.gz'
-        mutter("get_url %s" % url)
-        url_f = urllib2.urlopen(url)
-        if compressed:
-            return gzip.GzipFile(fileobj=StringIO(url_f.read()))
-        else:
-            return url_f
-
-
-
-def _find_remote_root(url):
-    """Return the prefix URL that corresponds to the branch root."""
-    orig_url = url
-    while True:
-        try:
-            fmt_url = url + '/.bzr/branch-format'
-            ff = get_url(fmt_url)
-            fmt = ff.read()
-            ff.close()
-
-            if fmt != BZR_BRANCH_FORMAT_5:
-                raise BzrError("sorry, branch format %r not supported at url %s"
-                               % (fmt, url))
-            
-            return url
-        except urllib2.URLError:
-            pass
-
-        scheme, host, path = list(urlparse.urlparse(url))[:3]
-        # discard params, query, fragment
-        
-        # strip off one component of the path component
-        idx = path.rfind('/')
-        if idx == -1 or path == '/':
-            raise BzrError('no branch root found for URL %s'
-                           ' or enclosing directories'
-                           % orig_url)
-        path = path[:idx]
-        url = urlparse.urlunparse((scheme, host, path, '', '', ''))
-        
-
-
-class RemoteBranch(LocalBranch):
-    def __init__(self, baseurl, find_root=True):
-        """Create new proxy for a remote branch."""
-        if find_root:
-            self.base = _find_remote_root(baseurl)
-        else:
-            self.base = baseurl
-        self._check_format(False)
-        # is guaranteed to be a v5 store
-
-        cfn = self.controlfilename
-        assert self._branch_format == 5
-        self.control_weaves = WeaveStore(cfn([]), get_url)
-        self.weave_store = WeaveStore(cfn('weaves'), get_url)
-        self.revision_store = RemoteStore(cfn('revision-store'))
-
-    def __str__(self):
-        b = getattr(self, 'baseurl', 'undefined')
-        return '%s(%r)' % (self.__class__.__name__, b)
-
-    __repr__ = __str__
-
-    def setup_caching(self, cache_root):
-        """Set up cached stores located under cache_root"""
-        from bzrlib.meta_store import CachedStore
-        for store_name in ('inventory_store', 'text_store', 'revision_store'):
-            if not isinstance(getattr(self, store_name), CachedStore):
-                cache_path = os.path.join(cache_root, store_name)
-                os.mkdir(cache_path)
-                new_store = CachedStore(getattr(self, store_name), cache_path)
-                setattr(self, store_name, new_store)
-
-    def controlfile(self, filename, mode):
-        if mode not in ('rb', 'rt', 'r'):
-            raise BzrError("file mode %r not supported for remote branches" % mode)
-        return get_url(self.base + '/.bzr/' + filename, False)
-
-
-    def lock_read(self):
-        # no locking for remote branches yet
-        pass
-
-    def lock_write(self):
-        from errors import LockError
-        raise LockError("write lock not supported for remote branch %s"
-                        % self.base)
-
-    def unlock(self):
-        pass
-    
-
-    def relpath(self, path):
-        if not path.startswith(self.base):
-            raise BzrError('path %r is not under base URL %r'
-                           % (path, self.base))
-        pl = len(self.base)
-        return path[pl:].lstrip('/')
-
-
-    def get_revision(self, revision_id):
-        try:
-            revf = self.revision_store[revision_id]
-        except KeyError:
-            raise NoSuchRevision(self, revision_id)
-        r = serializer_v5.read_revision(revf)
-        if r.revision_id != revision_id:
-            raise BzrCheckError('revision stored as {%s} actually contains {%s}'
-                                % (revision_id, r.revision_id))
-        return r
-
-
-class RemoteStore(object):
-    def __init__(self, baseurl):
-        self._baseurl = baseurl
-        
-
-    def _path(self, name):
-        if '/' in name:
-            raise ValueError('invalid store id', name)
-        return self._baseurl + '/' + name
-
-    def __contains__(self, fileid):
-        try:
-            self[fileid]
-            return True
-        except KeyError:
-            return False
-        
-    def __getitem__(self, fileid):
-        p = self._path(fileid)
-        try:
-            return get_url(p, compressed=True)
-        except urllib2.URLError:
-            pass
-        try:
-            return get_url(p, compressed=False)
-        except urllib2.URLError:
-            raise KeyError(fileid)
-    
-
-    

=== modified file 'bzrlib/selftest/__init__.py'
--- bzrlib/selftest/__init__.py	2005-09-29 12:35:37 +0000
+++ bzrlib/selftest/__init__.py	2005-09-29 20:34:25 +0000
@@ -462,6 +462,7 @@
                    'bzrlib.selftest.whitebox',
                    'bzrlib.selftest.teststore',
                    'bzrlib.selftest.blackbox',
+                   'bzrlib.selftest.testtransport',
                    'bzrlib.selftest.testgraph',
                    ]
 

=== modified file 'bzrlib/selftest/testremotebranch.py'
--- bzrlib/selftest/testremotebranch.py	2005-09-29 02:01:49 +0000
+++ bzrlib/selftest/testremotebranch.py	2005-09-29 20:34:25 +0000
@@ -25,10 +25,9 @@
 
 from bzrlib.selftest.testfetch import fetch_steps
 
-class TestFetch(TestCaseWithWebserver):
-    def runTest(self):
-        from bzrlib.fetch import greedy_fetch
-        from bzrlib.selftest.testfetch import has_revision
+class TestRemoteBranch(TestCaseWithWebserver):
+    def test_remote_branch(self):
+        from bzrlib.fetch import greedy_fetch, has_revision
 
         def new_branch(name):
             os.mkdir(name)

=== modified file 'bzrlib/selftest/teststore.py'
--- bzrlib/selftest/teststore.py	2005-09-28 07:08:18 +0000
+++ bzrlib/selftest/teststore.py	2005-09-29 20:34:25 +0000
@@ -16,32 +16,113 @@
 
 """Test Store implementation
 """
-
 from StringIO import StringIO
 import os
 
-from bzrlib.remotebranch import RemoteStore
-from bzrlib.store import ImmutableStore, copy_all
+from bzrlib.store import copy_all
+from bzrlib.transport.local import LocalTransport
+from bzrlib.transport import NoSuchFile
+from bzrlib.store.compressed_text import CompressedTextStore
 from bzrlib.selftest import TestCaseInTempDir
 from bzrlib.errors import BzrError, UnlistableStore
 
 
-class TestStore(TestCaseInTempDir):
+def fill_store(store):
+    store.add(StringIO('hello'), 'a')
+    store.add(StringIO('other'), 'b')
+    store.add(StringIO('something'), 'c')
+    store.add(StringIO('goodbye'), '123123')
+
+def check_equals(tester, store, files, values, permit_failure=False):
+    files = store.get(files, permit_failure=permit_failure)
+    count = 0
+    for f, v in zip(files, values):
+        count += 1
+        if v is None:
+            tester.assert_(f is None)
+        else:
+            tester.assertEquals(f.read(), v)
+    tester.assertEquals(count, len(values))
+    # We need to check to make sure there are no more
+    # files to be returned, I'm using a cheezy way
+    # Convert to a list, and there shouldn't be any left
+    tester.assertEquals(len(list(files)), 0)
+
+def test_multiple_add(tester, store):
+    fill_store(store)
+    tester.assertRaises(BzrError, store.add, StringIO('goodbye'), '123123')
+
+def test_get(tester, store):
+    fill_store(store)
+
+    check_equals(tester, store, ['a'], ['hello'])
+    check_equals(tester, store, ['b', 'c'], ['other', 'something'])
+
+    # Make sure that requesting a non-existing file fails
+    tester.assertRaises(NoSuchFile, check_equals, tester, store,
+            ['d'], [None])
+    tester.assertRaises(NoSuchFile, check_equals, tester, store,
+            ['a', 'd'], ['hello', None])
+    tester.assertRaises(NoSuchFile, check_equals, tester, store,
+            ['d', 'a'], [None, 'hello'])
+    tester.assertRaises(NoSuchFile, check_equals, tester, store,
+            ['d', 'd', 'd'], [None, None, None])
+    tester.assertRaises(NoSuchFile, check_equals, tester, store,
+            ['a', 'd', 'b'], ['hello', None, 'other'])
+
+def test_ignore_get(tester, store):
+    fill_store(store)
+
+    files = store.get(['d'], permit_failure=True)
+    files = list(files)
+    tester.assertEquals(len(files), 1)
+    tester.assert_(files[0] is None)
+
+    check_equals(tester, store, ['a', 'd'], ['hello', None],
+            permit_failure=True)
+    check_equals(tester, store, ['d', 'a'], [None, 'hello'],
+            permit_failure=True)
+    check_equals(tester, store, ['d', 'd'], [None, None],
+            permit_failure=True)
+    check_equals(tester, store, ['a', 'd', 'b'], ['hello', None, 'other'],
+            permit_failure=True)
+    check_equals(tester, store, ['a', 'd', 'b'], ['hello', None, 'other'],
+            permit_failure=True)
+    check_equals(tester, store, ['b', 'd', 'c'], ['other', None, 'something'],
+            permit_failure=True)
+
+def get_compressed_store():
+    t = LocalTransport('.')
+    return CompressedTextStore(t)
+
+class TestCompressedTextStore(TestCaseInTempDir):
     def test_multiple_add(self):
         """Multiple add with same ID should raise a BzrError"""
-        store = ImmutableStore('.')
-        store.add(StringIO('goodbye'), '123123')
-        self.assertRaises(BzrError, store.add, StringIO('goodbye'), '123123')
+        store = get_compressed_store()
+        test_multiple_add(self, store)
+
+    def test_get(self):
+        store = get_compressed_store()
+        test_get(self, store)
+
+    def test_ignore_get(self):
+        store = get_compressed_store()
+        test_ignore_get(self, store)
+
 
     def test_copy_all(self):
         """Test copying"""
         os.mkdir('a')
-        store_a = ImmutableStore('a')
+        store_a = CompressedTextStore('a')
         store_a.add('foo', '1')
         os.mkdir('b')
-        store_b = ImmutableStore('b')
+        store_b = CompressedTextStore('b')
         copy_all(store_a, store_b)
         self.assertEqual(store_a['1'].read(), 'foo')
         self.assertEqual(store_b['1'].read(), 'foo')
-        store_c = RemoteStore('http://example.com/')
-        self.assertRaises(UnlistableStore, copy_all, store_c, store_b)
+        # TODO: Switch the exception form UnlistableStore to
+        #       or make Stores throw UnlistableStore if their
+        #       Transport doesn't support listing
+        # store_c = RemoteStore('http://example.com/')
+        # self.assertRaises(UnlistableStore, copy_all, store_c, store_b)
+

=== added file 'bzrlib/selftest/testtransport.py'
--- bzrlib/selftest/testtransport.py	1970-01-01 00:00:00 +0000
+++ bzrlib/selftest/testtransport.py	2005-09-19 03:26:14 +0000
@@ -0,0 +1,515 @@
+# Copyright (C) 2004, 2005 by Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+
+import os
+from cStringIO import StringIO
+from bzrlib.selftest import TestCaseInTempDir
+from bzrlib.selftest.HTTPTestUtil import TestCaseWithWebserver
+from bzrlib.errors import NoSuchFile, FileExists, TransportNotPossible
+
+def _append(fn, txt):
+    """Append the given text (file-like object) to the supplied filename."""
+    f = open(fn, 'ab')
+    f.write(txt)
+    f.flush()
+    f.close()
+    del f
+
+class TestTransportMixIn(object):
+    """Subclass this, and it will provide a series of tests for a Transport.
+    It assumes that the Transport object is connected to the 
+    current working directory.  So that whatever is done 
+    through the transport, should show up in the working 
+    directory, and vice-versa.
+
+    This also tests to make sure that the functions work with both
+    generators and lists (assuming iter(list) is effectively a generator)
+    """
+    readonly = False
+    def get_transport(self):
+        """Children should override this to return the Transport object.
+        """
+        raise NotImplementedError
+
+    def test_has(self):
+        t = self.get_transport()
+
+        files = ['a', 'b', 'e', 'g']
+        self.build_tree(files)
+        self.assertEqual(t.has('a'), True)
+        self.assertEqual(t.has('c'), False)
+        self.assertEqual(list(t.has_multi(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'])),
+                [True, True, False, False, True, False, True, False])
+        self.assertEqual(list(t.has_multi(iter(['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']))),
+                [True, True, False, False, True, False, True, False])
+
+    def test_get(self):
+        t = self.get_transport()
+
+        files = ['a', 'b', 'e', 'g']
+        self.build_tree(files)
+        self.assertEqual(t.get('a').read(), open('a').read())
+        content_f = t.get_multi(files)
+        for path,f in zip(files, content_f):
+            self.assertEqual(open(path).read(), f.read())
+
+        content_f = t.get_multi(iter(files))
+        for path,f in zip(files, content_f):
+            self.assertEqual(open(path).read(), f.read())
+
+        self.assertRaises(NoSuchFile, t.get, 'c')
+        try:
+            files = list(t.get_multi(['a', 'b', 'c']))
+        except NoSuchFile:
+            pass
+        else:
+            self.fail('Failed to raise NoSuchFile for missing file in get_multi')
+        try:
+            files = list(t.get_multi(iter(['a', 'b', 'c', 'e'])))
+        except NoSuchFile:
+            pass
+        else:
+            self.fail('Failed to raise NoSuchFile for missing file in get_multi')
+
+    def test_put(self):
+        t = self.get_transport()
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.put, 'a', 'some text for a\n')
+            open('a', 'wb').write('some text for a\n')
+        else:
+            t.put('a', 'some text for a\n')
+        self.assert_(os.path.exists('a'))
+        self.check_file_contents('a', 'some text for a\n')
+        self.assertEqual(t.get('a').read(), 'some text for a\n')
+        # Make sure 'has' is updated
+        self.assertEqual(list(t.has_multi(['a', 'b', 'c', 'd', 'e'])),
+                [True, False, False, False, False])
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.put_multi,
+                    [('a', 'new\ncontents for\na\n'),
+                        ('d', 'contents\nfor d\n')])
+            open('a', 'wb').write('new\ncontents for\na\n')
+            open('d', 'wb').write('contents\nfor d\n')
+        else:
+            # Put also replaces contents
+            self.assertEqual(t.put_multi([('a', 'new\ncontents for\na\n'),
+                                          ('d', 'contents\nfor d\n')]),
+                             2)
+        self.assertEqual(list(t.has_multi(['a', 'b', 'c', 'd', 'e'])),
+                [True, False, False, True, False])
+        self.check_file_contents('a', 'new\ncontents for\na\n')
+        self.check_file_contents('d', 'contents\nfor d\n')
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                t.put_multi, iter([('a', 'diff\ncontents for\na\n'),
+                                  ('d', 'another contents\nfor d\n')]))
+            open('a', 'wb').write('diff\ncontents for\na\n')
+            open('d', 'wb').write('another contents\nfor d\n')
+        else:
+            self.assertEqual(
+                t.put_multi(iter([('a', 'diff\ncontents for\na\n'),
+                                  ('d', 'another contents\nfor d\n')]))
+                             , 2)
+        self.check_file_contents('a', 'diff\ncontents for\na\n')
+        self.check_file_contents('d', 'another contents\nfor d\n')
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.put, 'path/doesnt/exist/c', 'contents')
+        else:
+            self.assertRaises(NoSuchFile,
+                    t.put, 'path/doesnt/exist/c', 'contents')
+
+    def test_put_file(self):
+        t = self.get_transport()
+
+        # Test that StringIO can be used as a file-like object with put
+        f1 = StringIO('this is a string\nand some more stuff\n')
+        if self.readonly:
+            open('f1', 'wb').write(f1.read())
+        else:
+            t.put('f1', f1)
+
+        del f1
+
+        self.check_file_contents('f1', 
+                'this is a string\nand some more stuff\n')
+
+        f2 = StringIO('here is some text\nand a bit more\n')
+        f3 = StringIO('some text for the\nthird file created\n')
+
+        if self.readonly:
+            open('f2', 'wb').write(f2.read())
+            open('f3', 'wb').write(f3.read())
+        else:
+            t.put_multi([('f2', f2), ('f3', f3)])
+
+        del f2, f3
+
+        self.check_file_contents('f2', 'here is some text\nand a bit more\n')
+        self.check_file_contents('f3', 'some text for the\nthird file created\n')
+
+        # Test that an actual file object can be used with put
+        f4 = open('f1', 'rb')
+        if self.readonly:
+            open('f4', 'wb').write(f4.read())
+        else:
+            t.put('f4', f4)
+
+        del f4
+
+        self.check_file_contents('f4', 
+                'this is a string\nand some more stuff\n')
+
+        f5 = open('f2', 'rb')
+        f6 = open('f3', 'rb')
+        if self.readonly:
+            open('f5', 'wb').write(f5.read())
+            open('f6', 'wb').write(f6.read())
+        else:
+            t.put_multi([('f5', f5), ('f6', f6)])
+
+        del f5, f6
+
+        self.check_file_contents('f5', 'here is some text\nand a bit more\n')
+        self.check_file_contents('f6', 'some text for the\nthird file created\n')
+
+
+
+    def test_mkdir(self):
+        t = self.get_transport()
+
+        # Test mkdir
+        os.mkdir('dir_a')
+        self.assertEqual(t.has('dir_a'), True)
+        self.assertEqual(t.has('dir_b'), False)
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.mkdir, 'dir_b')
+            os.mkdir('dir_b')
+        else:
+            t.mkdir('dir_b')
+        self.assertEqual(t.has('dir_b'), True)
+        self.assert_(os.path.isdir('dir_b'))
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.mkdir_multi, ['dir_c', 'dir_d'])
+            os.mkdir('dir_c')
+            os.mkdir('dir_d')
+        else:
+            t.mkdir_multi(['dir_c', 'dir_d'])
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.mkdir_multi, iter(['dir_e', 'dir_f']))
+            os.mkdir('dir_e')
+            os.mkdir('dir_f')
+        else:
+            t.mkdir_multi(iter(['dir_e', 'dir_f']))
+        self.assertEqual(list(t.has_multi(
+            ['dir_a', 'dir_b', 'dir_c', 'dir_q',
+             'dir_d', 'dir_e', 'dir_f', 'dir_b'])),
+            [True, True, True, False,
+             True, True, True, True])
+        for d in ['dir_a', 'dir_b', 'dir_c', 'dir_d', 'dir_e', 'dir_f']:
+            self.assert_(os.path.isdir(d))
+
+        if not self.readonly:
+            self.assertRaises(NoSuchFile, t.mkdir, 'path/doesnt/exist')
+            self.assertRaises(FileExists, t.mkdir, 'dir_a') # Creating a directory again should fail
+
+        # Make sure the transport recognizes when a
+        # directory is created by other means
+        # Caching Transports will fail, because dir_e was already seen not
+        # to exist. So instead, we will search for a new directory
+        #os.mkdir('dir_e')
+        #if not self.readonly:
+        #    self.assertRaises(FileExists, t.mkdir, 'dir_e')
+
+        os.mkdir('dir_g')
+        if not self.readonly:
+            self.assertRaises(FileExists, t.mkdir, 'dir_g')
+
+        # Test get/put in sub-directories
+        if self.readonly:
+            open('dir_a/a', 'wb').write('contents of dir_a/a')
+            open('dir_b/b', 'wb').write('contents of dir_b/b')
+        else:
+            self.assertEqual(
+                t.put_multi([('dir_a/a', 'contents of dir_a/a'),
+                             ('dir_b/b', 'contents of dir_b/b')])
+                          , 2)
+        for f in ('dir_a/a', 'dir_b/b'):
+            self.assertEqual(t.get(f).read(), open(f).read())
+
+    def test_copy_to(self):
+        import tempfile
+        from bzrlib.transport.local import LocalTransport
+
+        t = self.get_transport()
+
+        files = ['a', 'b', 'c', 'd']
+        self.build_tree(files)
+
+        dtmp = tempfile.mkdtemp(dir='.', prefix='test-transport-')
+        dtmp_base = os.path.basename(dtmp)
+        local_t = LocalTransport(dtmp)
+
+        t.copy_to(files, local_t)
+        for f in files:
+            self.assertEquals(open(f).read(),
+                    open(os.path.join(dtmp_base, f)).read())
+
+        del dtmp, dtmp_base, local_t
+
+        dtmp = tempfile.mkdtemp(dir='.', prefix='test-transport-')
+        dtmp_base = os.path.basename(dtmp)
+        local_t = LocalTransport(dtmp)
+
+        files = ['a', 'b', 'c', 'd']
+        t.copy_to(iter(files), local_t)
+        for f in files:
+            self.assertEquals(open(f).read(),
+                    open(os.path.join(dtmp_base, f)).read())
+
+        del dtmp, dtmp_base, local_t
+
+    def test_append(self):
+        t = self.get_transport()
+
+        if self.readonly:
+            open('a', 'wb').write('diff\ncontents for\na\n')
+            open('b', 'wb').write('contents\nfor b\n')
+        else:
+            t.put_multi([
+                    ('a', 'diff\ncontents for\na\n'),
+                    ('b', 'contents\nfor b\n')
+                    ])
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.append, 'a', 'add\nsome\nmore\ncontents\n')
+            _append('a', 'add\nsome\nmore\ncontents\n')
+        else:
+            t.append('a', 'add\nsome\nmore\ncontents\n')
+
+        self.check_file_contents('a', 
+            'diff\ncontents for\na\nadd\nsome\nmore\ncontents\n')
+
+        if self.readonly:
+            self.assertRaises(TransportNotPossible,
+                    t.append_multi,
+                        [('a', 'and\nthen\nsome\nmore\n'),
+                         ('b', 'some\nmore\nfor\nb\n')])
+            _append('a', 'and\nthen\nsome\nmore\n')
+            _append('b', 'some\nmore\nfor\nb\n')
+        else:
+            t.append_multi([('a', 'and\nthen\nsome\nmore\n'),
+                    ('b', 'some\nmore\nfor\nb\n')])
+        self.check_file_contents('a', 
+            'diff\ncontents for\na\n'
+            'add\nsome\nmore\ncontents\n'
+            'and\nthen\nsome\nmore\n')
+        self.check_file_contents('b', 
+                'contents\nfor b\n'
+                'some\nmore\nfor\nb\n')
+
+        if self.readonly:
+            _append('a', 'a little bit more\n')
+            _append('b', 'from an iterator\n')
+        else:
+            t.append_multi(iter([('a', 'a little bit more\n'),
+                    ('b', 'from an iterator\n')]))
+        self.check_file_contents('a', 
+            'diff\ncontents for\na\n'
+            'add\nsome\nmore\ncontents\n'
+            'and\nthen\nsome\nmore\n'
+            'a little bit more\n')
+        self.check_file_contents('b', 
+                'contents\nfor b\n'
+                'some\nmore\nfor\nb\n'
+                'from an iterator\n')
+
+    def test_append_file(self):
+        t = self.get_transport()
+
+        contents = [
+            ('f1', 'this is a string\nand some more stuff\n'),
+            ('f2', 'here is some text\nand a bit more\n'),
+            ('f3', 'some text for the\nthird file created\n'),
+            ('f4', 'this is a string\nand some more stuff\n'),
+            ('f5', 'here is some text\nand a bit more\n'),
+            ('f6', 'some text for the\nthird file created\n')
+        ]
+        
+        if self.readonly:
+            for f, val in contents:
+                open(f, 'wb').write(val)
+        else:
+            t.put_multi(contents)
+
+        a1 = StringIO('appending to\none\n')
+        if self.readonly:
+            _append('f1', a1.read())
+        else:
+            t.append('f1', a1)
+
+        del a1
+
+        self.check_file_contents('f1', 
+                'this is a string\nand some more stuff\n'
+                'appending to\none\n')
+
+        a2 = StringIO('adding more\ntext to two\n')
+        a3 = StringIO('some garbage\nto put in three\n')
+
+        if self.readonly:
+            _append('f2', a2.read())
+            _append('f3', a3.read())
+        else:
+            t.append_multi([('f2', a2), ('f3', a3)])
+
+        del a2, a3
+
+        self.check_file_contents('f2',
+                'here is some text\nand a bit more\n'
+                'adding more\ntext to two\n')
+        self.check_file_contents('f3', 
+                'some text for the\nthird file created\n'
+                'some garbage\nto put in three\n')
+
+        # Test that an actual file object can be used with put
+        a4 = open('f1', 'rb')
+        if self.readonly:
+            _append('f4', a4.read())
+        else:
+            t.append('f4', a4)
+
+        del a4
+
+        self.check_file_contents('f4', 
+                'this is a string\nand some more stuff\n'
+                'this is a string\nand some more stuff\n'
+                'appending to\none\n')
+
+        a5 = open('f2', 'rb')
+        a6 = open('f3', 'rb')
+        if self.readonly:
+            _append('f5', a5.read())
+            _append('f6', a6.read())
+        else:
+            t.append_multi([('f5', a5), ('f6', a6)])
+
+        del a5, a6
+
+        self.check_file_contents('f5',
+                'here is some text\nand a bit more\n'
+                'here is some text\nand a bit more\n'
+                'adding more\ntext to two\n')
+        self.check_file_contents('f6',
+                'some text for the\nthird file created\n'
+                'some text for the\nthird file created\n'
+                'some garbage\nto put in three\n')
+
+    def test_get_partial(self):
+        t = self.get_transport()
+
+        contents = [
+            ('f1', 
+                'here is some text\nand a bit more\n'
+                'adding more\ntext to two\n'),
+            ('f2',
+                'this is a string\nand some more stuff\n'
+                'appending to\none\n'),
+            ('f3',
+                'some text for the\nthird file created\n'
+                'some garbage\nto put in three\n')
+        ]
+        if self.readonly:
+            for f, val in contents:
+                open(f, 'wb').write(val)
+        else:
+            t.put_multi(contents)
+
+        self.assertRaises(NoSuchFile,
+                t.get_partial, 'a-missing-file', 20)
+        self.assertRaises(NoSuchFile,
+                t.get_partial, 'another-missing-file', 20, 30)
+        f = t.get_partial('f1', 33)
+        self.assertEqual(f.read(), 
+                'adding more\ntext to two\n')
+        f = t.get_partial('f1', 33, 10)
+        self.assertEqual(f.read(10), 
+                'adding mor')
+
+        del f
+
+        offsets = [('f2', 37), ('f3', 20, 10), ('f1', 10, 20)]
+        values = ['appending to\none\n',
+                  'ird file c',
+                  'me text\nand a bit mo'
+                 ]
+        contents_f = t.get_partial_multi(offsets)
+        count = 0
+        for f, val in zip(contents_f, values):
+            count += 1
+            self.assertEqual(val, f.read(len(val)))
+        # Make sure we saw all values, and no extra
+        self.assertEqual(len(values), count)
+        self.assertEqual(list(contents_f), [])
+
+        # Do the same thing with an iterator
+        offsets = iter([('f2', 34), ('f3', 18, 10), ('f1', 15, 15)])
+        values = ['ff\nappending to\none\n',
+                  'third file',
+                  'xt\nand a bit mo'
+                 ]
+        contents_f = t.get_partial_multi(offsets)
+        count = 0
+        for f, val in zip(contents_f, values):
+            count += 1
+            self.assertEqual(val, f.read(len(val)))
+        self.assertEqual(len(values), count)
+        self.assertEqual(list(contents_f), [])
+
+
+    def test_delete(self):
+        # TODO: Test Transport.delete
+        pass
+
+    def test_move(self):
+        # TODO: Test Transport.move
+        pass
+
+class LocalTransportTest(TestCaseInTempDir, TestTransportMixIn):
+    def get_transport(self):
+        from bzrlib.transport.local import LocalTransport
+        return LocalTransport('.')
+
+class HttpTransportTest(TestCaseWithWebserver, TestTransportMixIn):
+    readonly = True
+    def get_transport(self):
+        from bzrlib.transport.http import HttpTransport
+        url = self.get_remote_url('.')
+        return HttpTransport(url)
+

=== added directory 'bzrlib/store'
=== renamed file 'bzrlib/store.py' => 'bzrlib/store/__init__.py'
--- bzrlib/store.py	2005-09-28 09:35:50 +0000
+++ bzrlib/store/__init__.py	2005-09-29 20:34:25 +0000
@@ -24,19 +24,9 @@
 unique ID.
 """
 
-import errno
-import gzip
-import os
-import tempfile
-import types
-from stat import ST_SIZE
-from StringIO import StringIO
-
 from bzrlib.errors import BzrError, UnlistableStore
 from bzrlib.trace import mutter
-import bzrlib.ui
-import bzrlib.osutils as osutils
-
+import bzrlib.transport
 
 ######################################################################
 # stores
@@ -44,230 +34,168 @@
 class StoreError(Exception):
     pass
 
-
-class ImmutableStore(object):
-    """Store that holds files indexed by unique names.
-
-    Files can be added, but not modified once they are in.  Typically
-    the hash is used as the name, or something else known to be unique,
-    such as a UUID.
-
-    >>> st = ImmutableScratchStore()
-
-    >>> st.add(StringIO('hello'), 'aa')
-    >>> 'aa' in st
-    True
-    >>> 'foo' in st
-    False
-
-    You are not allowed to add an id that is already present.
-
-    Entries can be retrieved as files, which may then be read.
-
-    >>> st.add(StringIO('goodbye'), '123123')
-    >>> st['123123'].read()
-    'goodbye'
-
-    TODO: Atomic add by writing to a temporary file and renaming.
-
-    In bzr 0.0.5 and earlier, files within the store were marked
-    readonly on disk.  This is no longer done but existing stores need
-    to be accomodated.
+class Store(object):
+    """This class represents the abstract storage layout for saving information.
     """
-
-    def __init__(self, basedir):
-        self._basedir = basedir
-
-    def _path(self, entry_id):
-        if not isinstance(entry_id, basestring):
-            raise TypeError(type(entry_id))
-        if '\\' in entry_id or '/' in entry_id:
-            raise ValueError("invalid store id %r" % entry_id)
-        return os.path.join(self._basedir, entry_id)
+    _transport = None
+    _max_buffered_requests = 10
+
+    def __init__(self, transport):
+        assert isinstance(transport, bzrlib.transport.Transport)
+        self._transport = transport
 
     def __repr__(self):
-        return "%s(%r)" % (self.__class__.__name__, self._basedir)
-
-    def add(self, f, fileid, compressed=True):
-        """Add contents of a file into the store.
-
-        f -- An open file, or file-like object."""
-        # FIXME: Only works on files that will fit in memory
-        
-        from bzrlib.atomicfile import AtomicFile
-        
-        mutter("add store entry %r" % (fileid))
-        if isinstance(f, types.StringTypes):
-            content = f
+        if self._transport is None:
+            return "%s(None)" % (self.__class__.__name__)
         else:
-            content = f.read()
-            
-        p = self._path(fileid)
-        if os.access(p, os.F_OK) or os.access(p + '.gz', os.F_OK):
-            raise BzrError("store %r already contains id %r" % (self._basedir, fileid))
-
-        fn = p
-        if compressed:
-            fn = fn + '.gz'
-            
-        af = AtomicFile(fn, 'wb')
-        try:
-            if compressed:
-                gf = gzip.GzipFile(mode='wb', fileobj=af)
-                gf.write(content)
-                gf.close()
+            return "%s(%r)" % (self.__class__.__name__, self._transport.base)
+
+    __str__ = __repr__
+
+    def __len__(self):
+        raise NotImplementedError('Children should define their length')
+
+    def __getitem__(self, fileid):
+        """Returns a file reading from a particular entry."""
+        raise NotImplementedError
+
+    def __contains__(self, fileid):
+        """"""
+        raise NotImplementedError
+
+    def __iter__(self):
+        raise NotImplementedError
+
+    def __del__(self):
+        """Delete an entry from the store."""
+        raise NotImplementedError('Children need to define delete rights.')
+
+    def add(self, f, fileid):
+        """Add a file object f to the store accessible from the given fileid"""
+        raise NotImplementedError('Children of Store must define their method of adding entries.')
+
+    def add_multi(self, entries):
+        """Add a series of file-like or string objects to the store with the given
+        identities.
+        
+        :param entries: A list of tuples of file,id pairs [(file1, id1), (file2, id2), ...]
+                        This could also be a generator yielding (file,id) pairs.
+        """
+        for f, fileid in entries:
+            self.add(f, fileid)
+
+    def has(self, fileids):
+        """Return True/False for each entry in fileids.
+
+        :param fileids: A List or generator yielding file ids.
+        :return: A generator or list returning True/False for each entry.
+        """
+        for fileid in fileids:
+            if fileid in self:
+                yield True
             else:
-                af.write(content)
-            af.commit()
-        finally:
-            af.close()
-
-
-    def copy_multi(self, other, ids, permit_failure=False):
+                yield False
+
+    def get(self, fileids, permit_failure=False, pb=None):
+        """Return a set of files, one for each requested entry.
+        
+        :param permit_failure: If true, return None for entries which do not 
+                               exist.
+        :return: A list or generator of file-like objects, one for each id.
+        """
+        for fileid in fileids:
+            try:
+                yield self[fileid]
+            except KeyError:
+                if permit_failure:
+                    yield None
+                else:
+                    raise
+
+    def copy_multi(self, other, ids, pb=None, permit_failure=False):
         """Copy texts for ids from other into self.
 
-        If an id is present in self, it is skipped.
+        If an id is present in self, it is skipped.  A count of copied
+        ids is returned, which may be less than len(ids).
 
-        Returns (count_copied, failed), where failed is a collection of ids
-        that could not be copied.
+        :param other: Another Store object
+        :param ids: A list of entry ids to be copied
+        :param pb: A ProgressBar object, if none is given, the default will be created.
+        :param permit_failure: Allow missing entries to be ignored
+        :return: (n_copied, [failed]) The number of entries copied successfully,
+            followed by a list of entries which could not be copied (because they
+            were missing)
         """
-        pb = bzrlib.ui.ui_factory.progress_bar()
-        
+        if pb is None:
+            pb = bzrlib.ui.ui_factory.progress_bar()
+
+        ids = list(ids) # Make sure we don't have a generator, since we iterate 2 times
         pb.update('preparing to copy')
-        to_copy = [id for id in ids if id not in self]
-        if isinstance(other, ImmutableStore):
-            return self.copy_multi_immutable(other, to_copy, pb, 
-                                             permit_failure=permit_failure)
-        count = 0
+        to_copy = []
+        for file_id, has in zip(ids, self.has(ids)):
+            if not has:
+                to_copy.append(file_id)
+        return self._do_copy(other, to_copy, pb, permit_failure=permit_failure)
+
+    def _do_copy(self, other, to_copy, pb, permit_failure=False):
+        """This is the standard copying mechanism, just get them one at
+        a time from remote, and store them locally.
+
+        :param other: Another Store object
+        :param to_copy: A list of entry ids to copy
+        :param pb: A ProgressBar object to display completion status.
+        :param permit_failure: Allow missing entries to be ignored
+        :return: (n_copied, [failed])
+            The number of entries copied, and a list of failed entries.
+        """
+        # This should be updated to use add_multi() rather than
+        # the current methods of buffering requests.
+        # One question, is it faster to queue up 1-10 and then copy 1-10
+        # then queue up 11-20, copy 11-20
+        # or to queue up 1-10, copy 1, queue 11, copy 2, etc?
+        # sort of pipeline versus batch.
+
+        # We can't use self._transport.copy_to because we don't know
+        # whether the local tree is in the same format as other
         failed = set()
-        for id in to_copy:
-            count += 1
-            pb.update('copy', count, len(to_copy))
-            if not permit_failure:
-                self.add(other[id], id)
-            else:
+        def buffer_requests():
+            count = 0
+            buffered_requests = []
+            for fileid in to_copy:
                 try:
-                    entry = other[id]
+                    f = other[fileid]
                 except KeyError:
-                    failed.add(id)
-                    continue
-                self.add(entry, id)
-                
-        if not permit_failure:
+                    if permit_failure:
+                        failed.add(fileid)
+                        continue
+                    else:
+                        raise
+
+                buffered_requests.append((f, fileid))
+                if len(buffered_requests) > self._max_buffered_requests:
+                    yield buffered_requests.pop(0)
+                    count += 1
+                    pb.update('copy', count, len(to_copy))
+
+            for req in buffered_requests:
+                yield req
+                count += 1
+                pb.update('copy', count, len(to_copy))
+
             assert count == len(to_copy)
-        pb.clear()
-        return count, failed
-
-    def copy_multi_immutable(self, other, to_copy, pb, permit_failure=False):
-        count = 0
-        failed = set()
-        for id in to_copy:
-            p = self._path(id)
-            other_p = other._path(id)
-            try:
-                osutils.link_or_copy(other_p, p)
-            except (IOError, OSError), e:
-                if e.errno == errno.ENOENT:
-                    if not permit_failure:
-                        osutils.link_or_copy(other_p+".gz", p+".gz")
-                    else:
-                        try:
-                            osutils.link_or_copy(other_p+".gz", p+".gz")
-                        except IOError, e:
-                            if e.errno == errno.ENOENT:
-                                failed.add(id)
-                            else:
-                                raise
-                else:
-                    raise
-            
-            count += 1
-            pb.update('copy', count, len(to_copy))
-        assert count == len(to_copy)
-        pb.clear()
-        return count, failed
-    
-
-    def __contains__(self, fileid):
-        """"""
-        p = self._path(fileid)
-        return (os.access(p, os.R_OK)
-                or os.access(p + '.gz', os.R_OK))
-
-    # TODO: Guard against the same thing being stored twice,
-    # compressed and uncompressed
-
-    def __iter__(self):
-        for f in os.listdir(self._basedir):
-            if f[-3:] == '.gz':
-                # TODO: case-insensitive?
-                yield f[:-3]
-            else:
-                yield f
-
-    def __len__(self):
-        return len(os.listdir(self._basedir))
-
-
-    def __getitem__(self, fileid):
-        """Returns a file reading from a particular entry."""
-        p = self._path(fileid)
-        try:
-            return gzip.GzipFile(p + '.gz', 'rb')
-        except IOError, e:
-            if e.errno != errno.ENOENT:
-                raise
-
-        try:
-            return file(p, 'rb')
-        except IOError, e:
-            if e.errno != errno.ENOENT:
-                raise
-
-        raise KeyError(fileid)
-
-
-    def total_size(self):
-        """Return (count, bytes)
-
-        This is the (compressed) size stored on disk, not the size of
-        the content."""
-        total = 0
-        count = 0
-        for fid in self:
-            count += 1
-            p = self._path(fid)
-            try:
-                total += os.stat(p)[ST_SIZE]
-            except OSError:
-                total += os.stat(p + '.gz')[ST_SIZE]
-                
-        return count, total
-
-
-
-
-class ImmutableScratchStore(ImmutableStore):
-    """Self-destructing test subclass of ImmutableStore.
-
-    The Store only exists for the lifetime of the Python object.
- Obviously you should not put anything precious in it.
-    """
-    def __init__(self):
-        ImmutableStore.__init__(self, tempfile.mkdtemp())
-
-    def __del__(self):
-        for f in os.listdir(self._basedir):
-            fpath = os.path.join(self._basedir, f)
-            # needed on windows, and maybe some other filesystems
-            os.chmod(fpath, 0600)
-            os.remove(fpath)
-        os.rmdir(self._basedir)
-        mutter("%r destroyed" % self)
+
+        self.add_multi(buffer_requests())
+
+        pb.clear()
+        return len(to_copy), failed
 
 def copy_all(store_from, store_to):
     """Copy all ids from one store to another."""
     if not hasattr(store_from, "__iter__"):
         raise UnlistableStore(store_from)
-    ids = [f for f in store_from]
+    try:
+        ids = [f for f in store_from]
+    except NotImplementedError:
+        raise UnlistableStore(store_from)
     store_to.copy_multi(store_from, ids)
+

=== added file 'bzrlib/store/compressed_text.py'
--- bzrlib/store/compressed_text.py	1970-01-01 00:00:00 +0000
+++ bzrlib/store/compressed_text.py	2005-09-17 21:57:11 +0000
@@ -0,0 +1,259 @@
+# Copyright (C) 2005 by Canonical Development Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""
+An implementation the primary storage type CompressedTextStore.
+
+This store keeps compressed versions of the full text. It does not
+do any sort of delta compression.
+"""
+
+import os, tempfile, gzip
+
+import bzrlib.store
+from bzrlib.trace import mutter
+from bzrlib.errors import BzrError
+
+from StringIO import StringIO
+from stat import ST_SIZE
+
+class CompressedTextStore(bzrlib.store.Store):
+    """Store that holds files indexed by unique names.
+
+    Files can be added, but not modified once they are in.  Typically
+    the hash is used as the name, or something else known to be unique,
+    such as a UUID.
+
+    Files are stored gzip compressed, with no delta compression.
+
+    >>> st = ScratchCompressedTextStore()
+
+    >>> st.add(StringIO('hello'), 'aa')
+    >>> 'aa' in st
+    True
+    >>> 'foo' in st
+    False
+
+    You are not allowed to add an id that is already present.
+
+    Entries can be retrieved as files, which may then be read.
+
+    >>> st.add(StringIO('goodbye'), '123123')
+    >>> st['123123'].read()
+    'goodbye'
+    """
+
+    def __init__(self, basedir):
+        super(CompressedTextStore, self).__init__(basedir)
+
+    def _check_fileid(self, fileid):
+        if '\\' in fileid or '/' in fileid:
+            raise ValueError("invalid store id %r" % fileid)
+
+    def _relpath(self, fileid):
+        self._check_fileid(fileid)
+        return fileid + '.gz'
+
+    def add(self, f, fileid):
+        """Add contents of a file into the store.
+
+        f -- An open file, or file-like object."""
+        # TODO: implement an add_multi which can do some of it's
+        #       own piplelining, and possible take advantage of
+        #       transport.put_multi(). The problem is that
+        #       entries potentially need to be compressed as they
+        #       are received, which implies translation, which
+        #       means it isn't as straightforward as we would like.
+        from cStringIO import StringIO
+        from bzrlib.osutils import pumpfile
+        
+        mutter("add store entry %r" % (fileid))
+        if isinstance(f, basestring):
+            f = StringIO(f)
+            
+        fn = self._relpath(fileid)
+        if self._transport.has(fn):
+            raise BzrError("store %r already contains id %r" % (self._transport.base, fileid))
+
+
+        sio = StringIO()
+        gf = gzip.GzipFile(mode='wb', fileobj=sio)
+        # if pumpfile handles files that don't fit in ram,
+        # so will this function
+        if isinstance(f, basestring):
+            gf.write(f)
+        else:
+            pumpfile(f, gf)
+        gf.close()
+        sio.seek(0)
+        self._transport.put(fn, sio)
+
+    def _do_copy(self, other, to_copy, pb, permit_failure=False):
+        if isinstance(other, CompressedTextStore):
+            return self._copy_multi_text(other, to_copy, pb,
+                    permit_failure=permit_failure)
+        return super(CompressedTextStore, self)._do_copy(other, to_copy,
+                pb, permit_failure=permit_failure)
+
+    def _copy_multi_text(self, other, to_copy, pb,
+            permit_failure=False):
+        # Because of _transport, we can no longer assume
+        # that they are on the same filesystem, we can, however
+        # assume that we only need to copy the exact bytes,
+        # we don't need to process the files.
+
+        failed = set()
+        if permit_failure:
+            new_to_copy = set()
+            for fileid, has in zip(to_copy, other.has(to_copy)):
+                if has:
+                    new_to_copy.add(fileid)
+                else:
+                    failed.add(fileid)
+            to_copy = new_to_copy
+            #mutter('_copy_multi_text copying %s, failed %s' % (to_copy, failed))
+
+        paths = [self._relpath(fileid) for fileid in to_copy]
+        count = other._transport.copy_to(paths, self._transport, pb=pb)
+        assert count == len(to_copy)
+        return count, failed
+
+    def __contains__(self, fileid):
+        """"""
+        fn = self._relpath(fileid)
+        return self._transport.has(fn)
+
+    def has(self, fileids, pb=None):
+        """Return True/False for each entry in fileids.
+
+        :param fileids: A List or generator yielding file ids.
+        :return: A generator or list returning True/False for each entry.
+        """
+        relpaths = (self._relpath(fid) for fid in fileids)
+        return self._transport.has_multi(relpaths, pb=pb)
+
+    def get(self, fileids, permit_failure=False, pb=None):
+        """Return a set of files, one for each requested entry.
+        
+        TODO: Write some tests to make sure that permit_failure is
+              handled correctly.
+
+        TODO: What should the exception be for a missing file?
+              KeyError, or NoSuchFile?
+        """
+
+        # This next code gets a bit hairy because it can allow
+        # to not request a file which doesn't seem to exist.
+        # Also, the same fileid may be requested twice, so we
+        # can't just build up a map.
+        rel_paths = [self._relpath(fid) for fid in fileids]
+        is_requested = []
+
+        #mutter('CompressedTextStore.get(permit_failure=%s)' % permit_failure)
+        if permit_failure:
+            existing_paths = []
+            for path, has in zip(rel_paths,
+                    self._transport.has_multi(rel_paths)):
+                if has:
+                    existing_paths.append(path)
+                    is_requested.append(True)
+                else:
+                    is_requested.append(False)
+            #mutter('Retrieving %s out of %s' % (existing_paths, rel_paths))
+        else:
+            #mutter('Retrieving all %s' % (rel_paths, ))
+            existing_paths = rel_paths
+            is_requested = [True for x in rel_paths]
+
+        count = 0
+        for f in self._transport.get_multi(existing_paths, pb=pb):
+            assert count < len(is_requested)
+            while not is_requested[count]:
+                yield None
+                count += 1
+            if hasattr(f, 'tell'):
+                yield gzip.GzipFile(mode='rb', fileobj=f)
+            else:
+                from cStringIO import StringIO
+                sio = StringIO(f.read())
+                yield gzip.GzipFile(mode='rb', fileobj=sio)
+            count += 1
+
+        while count < len(is_requested):
+            yield None
+            count += 1
+
+    def __iter__(self):
+        # TODO: case-insensitive?
+        for f in self._transport.list_dir('.'):
+            if f[-3:] == '.gz':
+                yield f[:-3]
+            else:
+                yield f
+
+    def __len__(self):
+        return len([f for f in self._transport.list_dir('.')])
+
+
+    def __getitem__(self, fileid):
+        """Returns a file reading from a particular entry."""
+        fn = self._relpath(fileid)
+        # This will throw if the file doesn't exist.
+        try:
+            f = self._transport.get(fn)
+        except:
+            raise KeyError('This store (%s) does not contain %s' % (self, fileid))
+
+        # gzip.GzipFile.read() requires a tell() function
+        # but some transports return objects that cannot seek
+        # so buffer them in a StringIO instead
+        if hasattr(f, 'tell'):
+            return gzip.GzipFile(mode='rb', fileobj=f)
+        else:
+            from cStringIO import StringIO
+            sio = StringIO(f.read())
+            return gzip.GzipFile(mode='rb', fileobj=sio)
+            
+
+    def total_size(self):
+        """Return (count, bytes)
+
+        This is the (compressed) size stored on disk, not the size of
+        the content."""
+        total = 0
+        count = 0
+        relpaths = [self._relpath(fid) for fid in self]
+        for st in self._transport.stat_multi(relpaths):
+            count += 1
+            total += st[ST_SIZE]
+                
+        return count, total
+
+class ScratchCompressedTextStore(CompressedTextStore):
+    """Self-destructing test subclass of CompressedTextStore.
+
+    The Store only exists for the lifetime of the Python object.
+    Obviously you should not put anything precious in it.
+    """
+    def __init__(self):
+        from transport import transport
+        super(ScratchCompressedTextStore, self).__init__(transport(tempfile.mkdtemp()))
+
+    def __del__(self):
+        self._transport.delete_multi(self._transport.list_dir('.'))
+        os.rmdir(self._transport.base)
+        mutter("%r destroyed" % self)
+

=== added directory 'bzrlib/transport'
=== added file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/__init__.py	2005-09-18 21:09:19 +0000
@@ -0,0 +1,355 @@
+# Copyright (C) 2005 Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+"""Transport is an abstraction layer to handle file access.
+
+The abstraction is to allow access from the local filesystem, as well
+as remote (such as http or sftp).
+"""
+
+from bzrlib.trace import mutter
+from bzrlib.errors import (BzrError, 
+    TransportError, TransportNotPossible, NonRelativePath,
+    NoSuchFile, FileExists, PermissionDenied,
+    ConnectionReset)
+
+_protocol_handlers = {
+}
+
+def register_transport(prefix, klass, override=True):
+    global _protocol_handlers
+
+    if _protocol_handlers.has_key(prefix):
+        if override:
+            mutter('overriding transport: %s => %s' % (prefix, klass.__name__))
+            _protocol_handlers[prefix] = klass
+    else:
+        mutter('registering transport: %s => %s' % (prefix, klass.__name__))
+        _protocol_handlers[prefix] = klass
+
+class Transport(object):
+    """This class encapsulates methods for retrieving or putting a file
+    from/to a storage location.
+
+    Most functions have a _multi variant, which allows you to queue up
+    multiple requests. They generally have a dumb base implementation 
+    which just iterates over the arguments, but smart Transport
+    implementations can do pipelining.
+    In general implementations should support having a generator or a list
+    as an argument (ie always iterate, never index)
+    """
+
+    def __init__(self, base):
+        super(Transport, self).__init__()
+        self.base = base
+
+    def clone(self, offset=None):
+        """Return a new Transport object, cloned from the current location,
+        using a subdirectory or parent directory. This allows connections 
+        to be pooled, rather than a new one needed for each subdir.
+        """
+        raise NotImplementedError
+
+    def should_cache(self):
+        """Return True if the data pulled across should be cached locally.
+        """
+        return False
+
+    def _pump(self, from_file, to_file):
+        """Most children will need to copy from one file-like 
+        object or string to another one.
+        This just gives them something easy to call.
+        """
+        if isinstance(from_file, basestring):
+            to_file.write(from_file)
+        else:
+            from bzrlib.osutils import pumpfile
+            pumpfile(from_file, to_file)
+
+    def _get_total(self, multi):
+        """Try to figure out how many entries are in multi,
+        but if not possible, return None.
+        """
+        try:
+            return len(multi)
+        except TypeError: # We can't tell how many, because relpaths is a generator
+            return None
+
+    def _update_pb(self, pb, msg, count, total):
+        """Update the progress bar based on the current count
+        and total available, total may be None if it was
+        not possible to determine.
+        """
+        if pb is None:
+            return
+        if total is None:
+            pb.update(msg, count, count+1)
+        else:
+            pb.update(msg, count, total)
+
+    def _iterate_over(self, multi, func, pb, msg, expand=True):
+        """Iterate over all entries in multi, passing them to func,
+        and update the progress bar as you go along.
+
+        :param expand:  If True, the entries will be passed to the function
+                        by expanding the tuple. If False, it will be passed
+                        as a single parameter.
+        """
+        total = self._get_total(multi)
+        count = 0
+        for entry in multi:
+            self._update_pb(pb, msg, count, total)
+            if expand:
+                func(*entry)
+            else:
+                func(entry)
+            count += 1
+        return count
+
+    def abspath(self, relpath):
+        """Return the full url to the given relative path.
+        This can be supplied with a string or a list
+        """
+        raise NotImplementedError
+
+    def relpath(self, abspath):
+        """Return the local path portion from a given absolute path.
+        """
+        raise NotImplementedError
+
+    def has(self, relpath):
+        """Does the target location exist?"""
+        raise NotImplementedError
+
+    def has_multi(self, relpaths, pb=None):
+        """Return True/False for each entry in relpaths"""
+        total = self._get_total(relpaths)
+        count = 0
+        for relpath in relpaths:
+            self._update_pb(pb, 'has', count, total)
+            yield self.has(relpath)
+            count += 1
+
+    def get(self, relpath):
+        """Get the file at the given relative path.
+
+        :param relpath: The relative path to the file
+        """
+        raise NotImplementedError
+
+    def get_multi(self, relpaths, pb=None):
+        """Get a list of file-like objects, one for each entry in relpaths.
+
+        :param relpaths: A list of relative paths.
+        :param pb:  An optional ProgressBar for indicating percent done.
+        :return: A list or generator of file-like objects
+        """
+        # TODO: Consider having this actually buffer the requests,
+        # in the default mode, it probably won't give worse performance,
+        # and all children wouldn't have to implement buffering
+        total = self._get_total(relpaths)
+        count = 0
+        for relpath in relpaths:
+            self._update_pb(pb, 'get', count, total)
+            yield self.get(relpath)
+            count += 1
+
+    def get_partial(self, relpath, start, length=None):
+        """Get just part of a file.
+
+        :param relpath: Path to the file, relative to base
+        :param start: The starting position to read from
+        :param length: The length to read. A length of None indicates
+                       read to the end of the file.
+        :return: A file-like object containing at least the specified bytes.
+                 Some implementations may return objects which can be read
+                 past this length, but this is not guaranteed.
+        """
+        raise NotImplementedError
+
+    def get_partial_multi(self, offsets, pb=None):
+        """Put a set of files or strings into the location.
+
+        Requesting multiple portions of the same file can be dangerous.
+
+        :param offsets: A list of tuples of relpath, start, length
+                         [(path1, start1, length1),
+                          (path2, start2, length2),
+                          (path3, start3), ...]
+                         length is optional, defaulting to None
+        :param pb:  An optional ProgressBar for indicating percent done.
+        :return: A generator of file-like objects.
+        """
+        total = self._get_total(offsets)
+        count = 0
+        for offset in offsets:
+            self._update_pb(pb, 'get_partial', count, total)
+            yield self.get_partial(*offset)
+            count += 1
+
+    def put(self, relpath, f):
+        """Copy the file-like or string object into the location.
+
+        :param relpath: Location to put the contents, relative to base.
+        :param f:       File-like or string object.
+        """
+        raise NotImplementedError
+
+    def put_multi(self, files, pb=None):
+        """Put a set of files or strings into the location.
+
+        :param files: A list of tuples of relpath, file object [(path1, file1), (path2, file2),...]
+        :param pb:  An optional ProgressBar for indicating percent done.
+        :return: The number of files copied.
+        """
+        return self._iterate_over(files, self.put, pb, 'put', expand=True)
+
+    def mkdir(self, relpath):
+        """Create a directory at the given path."""
+        raise NotImplementedError
+
+    def mkdir_multi(self, relpaths, pb=None):
+        """Create a group of directories"""
+        return self._iterate_over(relpaths, self.mkdir, pb, 'mkdir', expand=False)
+
+    def append(self, relpath, f):
+        """Append the text in the file-like or string object to 
+        the supplied location.
+        """
+        raise NotImplementedError
+
+    def append_multi(self, files, pb=None):
+        """Append the text in each file-like or string object to
+        the supplied location.
+
+        :param files: A set of (path, f) entries
+        :param pb:  An optional ProgressBar for indicating percent done.
+        """
+        return self._iterate_over(files, self.append, pb, 'append', expand=True)
+
+    def copy(self, rel_from, rel_to):
+        """Copy the item at rel_from to the location at rel_to"""
+        raise NotImplementedError
+
+    def copy_multi(self, relpaths, pb=None):
+        """Copy a bunch of entries.
+        
+        :param relpaths: A list of tuples of the form [(from, to), (from, to),...]
+        """
+        # This is the non-pipelined implementation, so that
+        # implementors don't have to implement everything.
+        return self._iterate_over(relpaths, self.copy, pb, 'copy', expand=True)
+
+    def copy_to(self, relpaths, other, pb=None):
+        """Copy a set of entries from self into another Transport.
+
+        :param relpaths: A list/generator of entries to be copied.
+        """
+        # The dummy implementation just does a simple get + put
+        def copy_entry(path):
+            other.put(path, self.get(path))
+
+        return self._iterate_over(relpaths, copy_entry, pb, 'copy_to', expand=False)
+
+
+    def move(self, rel_from, rel_to):
+        """Move the item at rel_from to the location at rel_to"""
+        raise NotImplementedError
+
+    def move_multi(self, relpaths, pb=None):
+        """Move a bunch of entries.
+        
+        :param relpaths: A list of tuples of the form [(from1, to1), (from2, to2),...]
+        """
+        return self._iterate_over(relpaths, self.move, pb, 'move', expand=True)
+
+    def move_multi_to(self, relpaths, rel_to):
+        """Move a bunch of entries to a single location.
+        This differs from move_multi in that you give a list of from, and
+        a single destination, rather than multiple destinations.
+
+        :param relpaths: A list of relative paths [from1, from2, from3, ...]
+        :param rel_to: A directory where each entry should be placed.
+        """
+        # This is not implemented, because you need to do special tricks to
+        # extract the basename, and add it to rel_to
+        raise NotImplementedError
+
+    def delete(self, relpath):
+        """Delete the item at relpath"""
+        raise NotImplementedError
+
+    def delete_multi(self, relpaths, pb=None):
+        """Queue up a bunch of deletes to be done.
+        """
+        return self._iterate_over(relpaths, self.delete, pb, 'delete', expand=False)
+
+    def stat(self, relpath):
+        """Return the stat information for a file.
+        WARNING: This may not be implementable for all protocols, so use
+        sparingly.
+        """
+        raise NotImplementedError
+
+    def stat_multi(self, relpaths, pb=None):
+        """Stat multiple files and return the information.
+        """
+        #TODO:  Is it worth making this a generator instead of a
+        #       returning a list?
+        stats = []
+        def gather(path):
+            stats.append(self.stat(path))
+
+        count = self._iterate_over(relpaths, gather, pb, 'stat', expand=False)
+        return stats
+
+
+    def list_dir(self, relpath):
+        """Return a list of all files at the given location.
+        WARNING: many transports do not support this, so trying avoid using
+        it if at all possible.
+        """
+        raise NotImplementedError
+
+    def lock_read(self, relpath):
+        """Lock the given file for shared (read) access.
+        WARNING: many transports do not support this, so trying avoid using it
+
+        :return: A lock object, which should contain an unlock() function.
+        """
+        raise NotImplementedError
+
+    def lock_write(self, relpath):
+        """Lock the given file for exclusive (write) access.
+        WARNING: many transports do not support this, so trying avoid using it
+
+        :return: A lock object, which should contain an unlock() function.
+        """
+        raise NotImplementedError
+
+
+def transport(base):
+    global _protocol_handlers
+    if base is None:
+        base = '.'
+    for proto, klass in _protocol_handlers.iteritems():
+        if proto is not None and base.startswith(proto):
+            return klass(base)
+    # The default handler is the filesystem handler
+    # which has a lookup of None
+    return _protocol_handlers[None](base)
+
+# Local transport should always be initialized
+import bzrlib.transport.local

=== added file 'bzrlib/transport/http.py'
--- bzrlib/transport/http.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/http.py	2005-09-18 21:09:19 +0000
@@ -0,0 +1,254 @@
+# Copyright (C) 2005 Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+"""Implementation of Transport over http.
+"""
+
+from bzrlib.transport import Transport, register_transport, \
+    TransportNotPossible, NoSuchFile, NonRelativePath, \
+    TransportError
+import os, errno
+from cStringIO import StringIO
+import urllib2
+import urlparse
+
+from bzrlib.errors import BzrError, BzrCheckError
+from bzrlib.branch import Branch, BZR_BRANCH_FORMAT
+from bzrlib.trace import mutter
+
+# velocitynet.com.au transparently proxies connections and thereby
+# breaks keep-alive -- sucks!
+
+
+def get_url(url):
+    import urllib2
+    mutter("get_url %s" % url)
+    url_f = urllib2.urlopen(url)
+    return url_f
+
+class HttpTransportError(TransportError):
+    pass
+
+class HttpTransport(Transport):
+    """This is the transport agent for http:// access.
+    
+    TODO: Implement pipelined versions of all of the *_multi() functions.
+    """
+
+    def __init__(self, base):
+        """Set the base path where files will be stored."""
+        assert base.startswith('http://') or base.startswith('https://')
+        super(HttpTransport, self).__init__(base)
+        # In the future we might actually connect to the remote host
+        # rather than using get_url
+        # self._connection = None
+        (self._proto, self._host,
+            self._path, self._parameters,
+            self._query, self._fragment) = urlparse.urlparse(self.base)
+
+    def should_cache(self):
+        """Return True if the data pulled across should be cached locally.
+        """
+        return True
+
+    def clone(self, offset=None):
+        """Return a new HttpTransport with root at self.base + offset
+        For now HttpTransport does not actually connect, so just return
+        a new HttpTransport object.
+        """
+        if offset is None:
+            return HttpTransport(self.base)
+        else:
+            return HttpTransport(self.abspath(offset))
+
+    def abspath(self, relpath):
+        """Return the full url to the given relative path.
+        This can be supplied with a string or a list
+        """
+        if isinstance(relpath, basestring):
+            relpath = [relpath]
+        basepath = self._path.split('/')
+        if len(basepath) > 0 and basepath[-1] == '':
+            basepath = basepath[:-1]
+
+        for p in relpath:
+            if p == '..':
+                if len(basepath) < 0:
+                    # In most filesystems, a request for the parent
+                    # of root, just returns root.
+                    continue
+                basepath.pop()
+            elif p == '.':
+                continue # No-op
+            else:
+                basepath.append(p)
+
+        # Possibly, we could use urlparse.urljoin() here, but
+        # I'm concerned about when it chooses to strip the last
+        # portion of the path, and when it doesn't.
+        path = '/'.join(basepath)
+        return urlparse.urlunparse((self._proto,
+                self._host, path, '', '', ''))
+
+    def relpath(self, abspath):
+        if not abspath.startswith(self.base):
+            raise NonRelativePath('path %r is not under base URL %r'
+                           % (abspath, self.base))
+        pl = len(self.base)
+        return abspath[pl:].lstrip('/')
+
+    def has(self, relpath):
+        """Does the target location exist?
+
+        TODO: HttpTransport.has() should use a HEAD request,
+        not a full GET request.
+
+        TODO: This should be changed so that we don't use
+        urllib2 and get an exception, the code path would be
+        cleaner if we just do an http HEAD request, and parse
+        the return code.
+        """
+        try:
+            f = get_url(self.abspath(relpath))
+            # Without the read and then close()
+            # we tend to have busy sockets.
+            f.read()
+            f.close()
+            return True
+        except BzrError:
+            return False
+        except urllib2.URLError:
+            return False
+        except IOError, e:
+            if e.errno == errno.ENOENT:
+                return False
+            raise HttpTransportError(orig_error=e)
+
+    def get(self, relpath, decode=False):
+        """Get the file at the given relative path.
+
+        :param relpath: The relative path to the file
+        """
+        try:
+            return get_url(self.abspath(relpath))
+        except (BzrError, urllib2.URLError, IOError), e:
+            raise NoSuchFile(orig_error=e)
+        except Exception,e:
+            raise HttpTransportError(orig_error=e)
+
+    def get_partial(self, relpath, start, length=None):
+        """Get just part of a file.
+
+        :param relpath: Path to the file, relative to base
+        :param start: The starting position to read from
+        :param length: The length to read. A length of None indicates
+                       read to the end of the file.
+        :return: A file-like object containing at least the specified bytes.
+                 Some implementations may return objects which can be read
+                 past this length, but this is not guaranteed.
+        """
+        # TODO: You can make specialized http requests for just
+        # a portion of the file. Figure out how to do that.
+        # For now, urllib2 returns files that cannot seek() so
+        # we just read bytes off the beginning, until we
+        # get to the point that we care about.
+        f = self.get(relpath)
+        # TODO: read in smaller chunks, in case things are
+        # buffered internally.
+        f.read(start)
+        return f
+
+    def put(self, relpath, f):
+        """Copy the file-like or string object into the location.
+
+        :param relpath: Location to put the contents, relative to base.
+        :param f:       File-like or string object.
+        """
+        raise TransportNotPossible('http PUT not supported')
+
+    def mkdir(self, relpath):
+        """Create a directory at the given path."""
+        raise TransportNotPossible('http does not support mkdir()')
+
+    def append(self, relpath, f):
+        """Append the text in the file-like object into the final
+        location.
+        """
+        raise TransportNotPossible('http does not support append()')
+
+    def copy(self, rel_from, rel_to):
+        """Copy the item at rel_from to the location at rel_to"""
+        raise TransportNotPossible('http does not support copy()')
+
+    def copy_to(self, relpaths, other, pb=None):
+        """Copy a set of entries from self into another Transport.
+
+        :param relpaths: A list/generator of entries to be copied.
+
+        TODO: if other is LocalTransport, is it possible to
+              do better than put(get())?
+        """
+        # At this point HttpTransport might be able to check and see if
+        # the remote location is the same, and rather than download, and
+        # then upload, it could just issue a remote copy_this command.
+        if isinstance(other, HttpTransport):
+            raise TransportNotPossible('http cannot be the target of copy_to()')
+        else:
+            return super(HttpTransport, self).copy_to(relpaths, other, pb=pb)
+
+    def move(self, rel_from, rel_to):
+        """Move the item at rel_from to the location at rel_to"""
+        raise TransportNotPossible('http does not support move()')
+
+    def delete(self, relpath):
+        """Delete the item at relpath"""
+        raise TransportNotPossible('http does not support delete()')
+
+    def list_dir(self, relpath):
+        """Return a list of all files at the given location.
+        WARNING: many transports do not support this, so trying avoid using
+        it if at all possible.
+        """
+        raise TransportNotPossible('http does not support list_dir()')
+
+    def stat(self, relpath):
+        """Return the stat information for a file.
+        """
+        raise TransportNotPossible('http does not support stat()')
+
+    def lock_read(self, relpath):
+        """Lock the given file for shared (read) access.
+        :return: A lock object, which should be passed to Transport.unlock()
+        """
+        # The old RemoteBranch ignore lock for reading, so we will
+        # continue that tradition and return a bogus lock object.
+        class BogusLock(object):
+            def __init__(self, path):
+                self.path = path
+            def unlock(self):
+                pass
+        return BogusLock(relpath)
+
+    def lock_write(self, relpath):
+        """Lock the given file for exclusive (write) access.
+        WARNING: many transports do not support this, so trying avoid using it
+
+        :return: A lock object, which should be passed to Transport.unlock()
+        """
+        raise TransportNotPossible('http does not support lock_write()')
+
+register_transport('http://', HttpTransport)
+register_transport('https://', HttpTransport)
+

=== added file 'bzrlib/transport/local.py'
--- bzrlib/transport/local.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/local.py	2005-09-18 21:09:19 +0000
@@ -0,0 +1,226 @@
+# Copyright (C) 2005 Canonical Ltd
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+"""Implementation of Transport for the local filesystem.
+"""
+
+from bzrlib.transport import Transport, register_transport, \
+    TransportError, NoSuchFile, FileExists
+import os, errno
+
+class LocalTransportError(TransportError):
+    pass
+
+class LocalTransport(Transport):
+    """This is the transport agent for local filesystem access."""
+
+    def __init__(self, base):
+        """Set the base path where files will be stored."""
+        if base.startswith('file://'):
+            base = base[7:]
+        super(LocalTransport, self).__init__(os.path.realpath(base))
+
+    def should_cache(self):
+        return False
+
+    def clone(self, offset=None):
+        """Return a new LocalTransport with root at self.base + offset
+        Because the local filesystem does not require a connection, 
+        we can just return a new object.
+        """
+        if offset is None:
+            return LocalTransport(self.base)
+        else:
+            return LocalTransport(self.abspath(offset))
+
+    def abspath(self, relpath):
+        """Return the full url to the given relative path.
+        This can be supplied with a string or a list
+        """
+        if isinstance(relpath, basestring):
+            relpath = [relpath]
+        return os.path.join(self.base, *relpath)
+
+    def relpath(self, abspath):
+        """Return the local path portion from a given absolute path.
+        """
+        from bzrlib.branch import _relpath
+        return _relpath(self.base, abspath)
+
+    def has(self, relpath):
+        return os.access(self.abspath(relpath), os.F_OK)
+
+    def get(self, relpath):
+        """Get the file at the given relative path.
+
+        :param relpath: The relative path to the file
+        """
+        try:
+            path = self.abspath(relpath)
+            return open(path, 'rb')
+        except IOError,e:
+            if e.errno == errno.ENOENT:
+                raise NoSuchFile('File %r does not exist' % path, orig_error=e)
+            raise LocalTransportError(orig_error=e)
+
+    def get_partial(self, relpath, start, length=None):
+        """Get just part of a file.
+
+        :param relpath: Path to the file, relative to base
+        :param start: The starting position to read from
+        :param length: The length to read. A length of None indicates
+                       read to the end of the file.
+        :return: A file-like object containing at least the specified bytes.
+                 Some implementations may return objects which can be read
+                 past this length, but this is not guaranteed.
+        """
+        # LocalTransport.get_partial() doesn't care about the length
+        # argument, because it is using a local file, and thus just
+        # returns the file seek'ed to the appropriate location.
+        try:
+            path = self.abspath(relpath)
+            f = open(path, 'rb')
+            f.seek(start, 0)
+            return f
+        except IOError,e:
+            if e.errno == errno.ENOENT:
+                raise NoSuchFile('File %r does not exist' % path, orig_error=e)
+            raise LocalTransportError(orig_error=e)
+
+    def put(self, relpath, f):
+        """Copy the file-like or string object into the location.
+
+        :param relpath: Location to put the contents, relative to base.
+        :param f:       File-like or string object.
+        """
+        from bzrlib.atomicfile import AtomicFile
+
+        try:
+            path = self.abspath(relpath)
+            fp = AtomicFile(path, 'wb')
+        except IOError, e:
+            if e.errno == errno.ENOENT:
+                raise NoSuchFile('File %r does not exist' % path, orig_error=e)
+            raise LocalTransportError(orig_error=e)
+        try:
+            self._pump(f, fp)
+            fp.commit()
+        finally:
+            fp.close()
+
+    def mkdir(self, relpath):
+        """Create a directory at the given path."""
+        try:
+            os.mkdir(self.abspath(relpath))
+        except OSError,e:
+            if e.errno == errno.EEXIST:
+                raise FileExists(orig_error=e)
+            elif e.errno == errno.ENOENT:
+                raise NoSuchFile(orig_error=e)
+            raise LocalTransportError(orig_error=e)
+
+    def append(self, relpath, f):
+        """Append the text in the file-like object into the final
+        location.
+        """
+        fp = open(self.abspath(relpath), 'ab')
+        self._pump(f, fp)
+
+    def copy(self, rel_from, rel_to):
+        """Copy the item at rel_from to the location at rel_to"""
+        import shutil
+        path_from = self.abspath(rel_from)
+        path_to = self.abspath(rel_to)
+        try:
+            shutil.copy(path_from, path_to)
+        except OSError,e:
+            raise LocalTransportError(orig_error=e)
+
+    def move(self, rel_from, rel_to):
+        """Move the item at rel_from to the location at rel_to"""
+        path_from = self.abspath(rel_from)
+        path_to = self.abspath(rel_to)
+
+        try:
+            os.rename(path_from, path_to)
+        except OSError,e:
+            raise LocalTransportError(orig_error=e)
+
+    def delete(self, relpath):
+        """Delete the item at relpath"""
+        try:
+            os.remove(self.abspath(relpath))
+        except OSError,e:
+            raise LocalTransportError(orig_error=e)
+
+    def copy_to(self, relpaths, other, pb=None):
+        """Copy a set of entries from self into another Transport.
+
+        :param relpaths: A list/generator of entries to be copied.
+        """
+        if isinstance(other, LocalTransport):
+            # Both from & to are on the local filesystem
+            # Unfortunately, I can't think of anything faster than just
+            # copying them across, one by one :(
+            import shutil
+
+            total = self._get_total(relpaths)
+            count = 0
+            for path in relpaths:
+                self._update_pb(pb, 'copy-to', count, total)
+                shutil.copy(self.abspath(path), other.abspath(path))
+                count += 1
+            return count
+        else:
+            return super(LocalTransport, self).copy_to(relpaths, other, pb=pb)
+
+
+    def list_dir(self, relpath):
+        """Return a list of all files at the given location.
+        WARNING: many transports do not support this, so trying avoid using
+        it if at all possible.
+        """
+        try:
+            return os.listdir(self.abspath(relpath))
+        except OSError,e:
+            raise LocalTransportError(orig_error=e)
+
+    def stat(self, relpath):
+        """Return the stat information for a file.
+        """
+        try:
+            return os.stat(self.abspath(relpath))
+        except OSError,e:
+            raise LocalTransportError(orig_error=e)
+
+    def lock_read(self, relpath):
+        """Lock the given file for shared (read) access.
+        :return: A lock object, which should be passed to Transport.unlock()
+        """
+        from bzrlib.lock import ReadLock
+        return ReadLock(self.abspath(relpath))
+
+    def lock_write(self, relpath):
+        """Lock the given file for exclusive (write) access.
+        WARNING: many transports do not support this, so trying avoid using it
+
+        :return: A lock object, which should be passed to Transport.unlock()
+        """
+        from bzrlib.lock import WriteLock
+        return WriteLock(self.abspath(relpath))
+
+# If nothing else matches, try the LocalTransport
+register_transport(None, LocalTransport)
+register_transport('file://', LocalTransport)

